'From Squeak 3.2 of 11 July 2002 [latest update: #4917] on 7 March 2003 at 1:17:13 pm'!
AlignmentMorph subclass: #GoGoBoardMorph
	instanceVariableNames: 'portNum port readouts sensorValues currentState highByte '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scratch'!
!GoGoBoardMorph commentStamp: '<historical>' prior: 0!
I represent the Squeak interface to a GoGo board connected via a serial port. The GoGo board has 8 10-bit sensor inputs capable of reading either voltage or resistive sensors plus three output ports with sufficient current to control DC motors, small lamps, or similar devices. See http://learning.media.mit.edu/projects/gogo/ for additional info about the GoGo board.

To create an instance of me:

	GoGoBoardMorph new openInWorld

(Note: You can only have one instance of me for each serial port that has a GoGo Board attached.)

Use the "Port?" button to choose the serial port. Then click "Start" to start reading sensor data. To control motors, just use my motor control commands. When you're done, click the "Stop" button.
!


!GoGoBoardMorph methodsFor: 'initialization' stamp: 'jm 3/5/2003 22:12'!
addButtons
	"Add buttons to choose a port and to start or stop the GoGo Board."

	| b r |
	b _ SimpleButtonMorph new target: self; borderColor: Color black.
	r _ AlignmentMorph newRow.
	r color: self color; borderWidth: 0; inset: 2.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	r centering: #topLeft.
	r addMorphBack: (b fullCopy label: 'Port?';	actionSelector: #selectPort).
	r addMorphBack: (b fullCopy label: 'Start';	actionSelector: #startReadingData).
	r addMorphBack: (b fullCopy label: 'Stop';	actionSelector: #stopReadingData).
	self addMorphFront: r.

! !

!GoGoBoardMorph methodsFor: 'initialization' stamp: 'jm 3/5/2003 22:13'!
addReadoutLabeled: aString
	"Add a readout row with the given label. Answer the stringmorph that's the readout."

	| row box readout |
	row _ AlignmentMorph newRow color: color.
	row addMorphBack: (StringMorph contents: aString).
	row addMorphBack: (Morph new color: color; extent: 10@5). "spacer"
	box _ AlignmentMorph newColumn
		extent: 80@18;
		hResizing: #rigid;
		vResizing: #rigid;
		borderWidth: 1;
		centering: #center;
		color: Color veryLightGray.
	readout _ StringMorph contents: '0'.
	box addMorph: readout.
	row addMorphBack: box.
	self addMorphBack: row.
	
	^ readout
! !

!GoGoBoardMorph methodsFor: 'initialization' stamp: 'jm 3/7/2003 13:16'!
initialize

	super initialize.
	self orientation: #vertical.
	self hResizing: #shrinkWrap.
	self vResizing: #shrinkWrap.
	self color: (Color r: 0.9 g: 1.0 b: 0.9).
	self borderWidth: 2.
	self inset: 3.
	self addButtons.

	portNum _ 1.
	readouts _ (1 to: 8) collect: [:i | self addReadoutLabeled: i printString].
	sensorValues _ Array new: 8 withAll: 0.
	currentState _ #idle.
	highByte _ 0.
! !


!GoGoBoardMorph methodsFor: 'stepping' stamp: 'jm 3/5/2003 21:36'!
step

	self portIsOpen ifFalse: [port _ nil. ^ self].
	self processIncomingData.
	1 to: 8 do: [:i |
		(readouts at: i) contents: (sensorValues at: i) printString].
! !

!GoGoBoardMorph methodsFor: 'stepping' stamp: 'jm 3/5/2003 19:41'!
stepTime

	^ 100
! !


!GoGoBoardMorph methodsFor: 'serial port' stamp: 'jm 3/5/2003 21:04'!
closePort

	port ifNotNil: [port close].
	port _ nil.
! !

!GoGoBoardMorph methodsFor: 'serial port' stamp: 'jm 3/5/2003 21:12'!
openPort: anInteger

	self closePort.
	port _ SerialPort new openPort: anInteger.
	currentState _ #idle.
	sensorValues _ Array new: 8 withAll: 0.
	self startStreamingSensors: 0.  "in case board was left in streaming mode"
	(Delay forMilliseconds: 100) wait.
	port flushInputBuffer.
	self ping.
! !

!GoGoBoardMorph methodsFor: 'serial port' stamp: 'jm 3/5/2003 21:35'!
portIsOpen

	^ port notNil and: [port isOpen]
! !

!GoGoBoardMorph methodsFor: 'serial port' stamp: 'jm 3/7/2003 13:14'!
selectPort

	| menu n |
	menu _ CustomMenu new.
	(0 to: 8) do: [:w | menu add: w printString action: w].
	n _ menu startUp.
	n ifNotNil: [portNum _ n].
! !


!GoGoBoardMorph methodsFor: 'sensor ops' stamp: 'jm 3/5/2003 19:54'!
ping
	"Answer true if a GoGoBoard is connected to my serial port."

	| buf |
	port flushInputBuffer.
	port nextPutAll: #(84 254 0) asByteArray.
	buf _ port next: 3.
	^ buf = #(85 255 170) asByteArray
! !

!GoGoBoardMorph methodsFor: 'sensor ops' stamp: 'jm 3/5/2003 21:23'!
startReadingData

	self closePort.
	self openPort: portNum.
	self startStreamingSensors: 16rFF.  "all 8 sensors"
! !

!GoGoBoardMorph methodsFor: 'sensor ops' stamp: 'jm 3/5/2003 19:54'!
startStreamingSensors: sensorByte
	"Begin streaming data from the set of sensors specified by the bits of the given byte. Invoke this method with 0 to stop streaming. While the data is streaming, data can be read using repeated calls similar such as 'gogoBoard port next: 3'. The data value is the least significant two bits of the second byte and the eight bits of the third byte of each 3-byte sensor reading. 84 is 16r54 in base 10."

	| cmd |
	port flushInputBuffer.
	cmd _ #(84 254 160 0) asByteArray.
	cmd at: 4 put: sensorByte.
	port nextPutAll: cmd.
	port next: 3.  "consume the 3-byte repy to this command"
! !

!GoGoBoardMorph methodsFor: 'sensor ops' stamp: 'jm 3/5/2003 21:25'!
stopReadingData

	(port notNil and: [port isOpen]) ifTrue: [self startStreamingSensors: 0].  "turn off streaming"
	self closePort.

! !


!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 21:52'!
coastToStopMotor: motorNum
	"Let the given motor coast to a stop."

	self portIsOpen ifFalse: [^ self].
	self selectMotor: motorNum.
	port nextPutAll: #(84 254 84) asByteArray.
! !

!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 21:50'!
reverseMotor: motorNum
	"Reverse the direction of the given motor."

	self portIsOpen ifFalse: [^ self].
	self selectMotor: motorNum.
	port nextPutAll: #(84 254 72) asByteArray.
! !

!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 22:53'!
selectMotor: anInteger
	"Select the motor to which subsequent motor commands will be addressed. Motors are numbered 1-6."

	| motorNum msg |
	motorNum _ (anInteger asInteger max: 1) min: 6.
	msg _ #(84 254 128 0) asByteArray.
	msg at: 4 put: (1 bitShift: (motorNum - 1)).
	port nextPutAll: msg.

! !

!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 23:00'!
setPower: aNumber motor: motorNum
	"Set the power of the given motor to 1 to 7."

	| power msg |
	self portIsOpen ifFalse: [^ self].
	self selectMotor: motorNum.
	power _ (aNumber truncated min: 7) max: 1.
	msg _ #(84 254 0) asByteArray.
	msg at: 3 put: 96 + ((power - 1) * 4).
	port nextPutAll: msg.
! !

!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 21:46'!
thatWayMotor: motorNum
	"Set motor direction to that way."

	self portIsOpen ifFalse: [^ self].
	self selectMotor: motorNum.
	port nextPutAll: #(84 254 80) asByteArray.
! !

!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 21:49'!
thisWayMotor: motorNum
	"Set motor direction to this way."

	self portIsOpen ifFalse: [^ self].
	self selectMotor: motorNum.
	port nextPutAll: #(84 254 76) asByteArray.
! !

!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 22:01'!
turnOffMotor: motorNum

	self portIsOpen ifFalse: [^ self].
	self selectMotor: motorNum.
	port nextPutAll: #(84 254 68) asByteArray.

! !

!GoGoBoardMorph methodsFor: 'motor ops' stamp: 'jm 3/5/2003 22:01'!
turnOnMotor: motorNum

	self portIsOpen ifFalse: [^ self].
	self selectMotor: motorNum.
	port nextPutAll: #(84 254 64) asByteArray.

! !


!GoGoBoardMorph methodsFor: 'private' stamp: 'jm 3/5/2003 21:03'!
processByte: aByte
	"Process one byte of the incoming data stream."
	"Details: This code recognizes three-byte sensor update messages starting with 16r0C using a simple, three-state finite state machine. We assume that sensor updates are always contiguous--that is, replies to motor or other GoGo board commands will not be inserted between the bytes of any given three-byte sensor update message. The sensor number is the top three bits of the byte two. The data value is the bottom two bits of byte two plus all eight bits of byte three."

	| sensorNum val |
	currentState = #idle ifTrue: [ 
		aByte = 16r0C ifTrue: [currentState _ #startByteSeen].
		^ self].

	currentState = #startByteSeen ifTrue: [
		highByte _ aByte.
		currentState _ #highByteSeen.
		^ self].

	currentState = #highByteSeen ifTrue: [
		"final byte of message: report the sensor value"
		sensorNum _ highByte bitShift: -5.
		val _ ((highByte bitAnd: 3) bitShift: 8) + aByte.
		sensorValues at: sensorNum + 1 put: val.

		currentState _ #idle].
! !

!GoGoBoardMorph methodsFor: 'private' stamp: 'jm 3/5/2003 22:35'!
processIncomingData
	"Process incoming bytes from the serial port."
	"On a Macintosh G4 Powerbook, I measured a sensor reading rate of 234 sensor readings/sec."

	| buf |
	buf _ port readByteArray.
	buf do: [:b | self processByte: b].
! !
